from zope.container.constraints import contains
from zope.container.interfaces import IContainer
from zope.interface import Attribute
from zope.interface.common.sequence import IFiniteSequence
from zope.location.interfaces import IContained


class IKeyManager(IContainer):
    contains("plone.keyring.interfaces.IKeyring")

    def clear(ring="_system"):
        """Clear all keys on a given ring. By default the system ring
        is cleader.  If None is used as ring id all rings are cleared.
        """

    def rotate(ring="_system"):
        """Rotate a given ring. By default rotates the system ring.
        If None is used as ring id all rings are rotated.
        """

    def secret(ring="_system"):
        """Return the current secret for a given ring. If no ring
        is given the secret for the system ring is returned"""


class IKeyring(IContained, IFiniteSequence):
    current = Attribute("The current (ie latest) secret in the ring.")

    def __init__(size=5):
        """Construct a new keyring for a specified number of keys."""

    def clear():
        """Remove all keys from the ring."""

    def rotate():
        """Add a new secret to the ring, pushing out the oldest secret."""



from plone.keyring.interfaces import IKeyring
from plone.keyring.keyring import Keyring
from unittest import defaultTestLoader
from unittest import TestCase
from unittest import TestSuite
from zope.interface.verify import verifyClass

import types


class KeyringTests(TestCase):
    def testInterface(self):
        verifyClass(IKeyring, Keyring)

    def testConstructionDefaultSize(self):
        ring = Keyring()
        self.assertEqual(len(ring), 5)

    def testConstructionSize(self):
        ring = Keyring(3)
        self.assertEqual(len(ring), 3)

    def testKeyringStartsEmpty(self):
        ring = Keyring()
        self.assertEqual(set(list(ring)), {None})

    def testIterate(self):
        ring = Keyring()
        ring.data = [0, 1, 2, 3, 4]
        iterator = ring.__iter__()
        self.assertTrue(isinstance(iterator, types.GeneratorType))
        self.assertEqual(list(iterator), [0, 1, 2, 3, 4])

    def testClear(self):
        ring = Keyring()
        ring.data = [0, 1, 2]
        ring.clear()
        self.assertEqual(ring.data, [None, None, None])

    def testRotate(self):
        ring = Keyring()
        ring.rotate()
        self.assertFalse(ring.current is None)
        self.assertEqual(ring.data[1:], [None, None, None, None])

    def testRotateTwice(self):
        ring = Keyring()
        ring.rotate()
        ring.rotate()
        self.assertTrue(ring.data[0] is not None)
        self.assertTrue(ring.data[1] is not None)
        self.assertEqual(ring.data[2:], [None, None, None])

    def testCurrent(self):
        ring = Keyring()
        marker = []
        ring.data = [marker, 1, 2, 3]
        self.assertTrue(ring.current is marker)


def test_suite():
    suite = TestSuite()
    suite.addTest(defaultTestLoader.loadTestsFromTestCase(KeyringTests))
    return suite



from persistent.list import PersistentList
from plone.keyring import django_random
from plone.keyring.interfaces import IKeyring
from random import choice
from zope.interface import implementer
from zope.location.interfaces import IContained

import time


def GenerateSecret(length=64):
    return django_random.get_random_string(length)


@implementer(IKeyring, IContained)
class Keyring(PersistentList):
    __parent__ = __name__ = None

    last_rotation = 0

    def __init__(self, size=5):
        PersistentList.__init__(self)
        for i in range(size):
            self.append(None)

    def __iter__(self):
        yield from self.data

    def clear(self):
        for i in range(len(self)):
            self[i] = None

    def rotate(self):
        self.pop()
        self.insert(0, GenerateSecret())
        self.last_rotation = time.time()

    def fill(self):
        """
        add missing keys
        """
        for i in range(len(self)):
            key = self[i]
            if key is None:
                self[i] = GenerateSecret()

    @property
    def current(self):
        return self.data[0]

    def random(self):
        """
        since we could be on a rotation boundary,
        only rotate one less than the total
        """
        keys = self.data
        if len(keys) > 1:
            keys = keys[:-1]
        return choice([k for k in keys if k])



from persistent.mapping import PersistentMapping
from plone.keyring.interfaces import IKeyManager
from plone.keyring.keymanager import KeyManager
from plone.keyring.keyring import Keyring
from unittest import defaultTestLoader
from unittest import TestCase
from unittest import TestSuite
from zope.interface.verify import verifyClass


marker = []


class KeyManagerTests(TestCase):
    def setUp(self):
        self.mgr = KeyManager()
        del self.mgr["_system"]
        self.mgr["_system"] = Keyring()
        self.mgr["_system"].rotate()
        self.mgr["one"] = Keyring()
        self.mgr["one"].rotate()
        self.mgr["two"] = Keyring()
        self.mgr["two"].rotate()

    def testInterface(self):
        verifyClass(IKeyManager, KeyManager)

    def testSystemKeyringCreated(self):
        mgr = KeyManager()
        self.assertEqual(set(mgr), {"_anon", "_forms", "_system"})
        self.assertTrue(mgr["_system"].current is not None)

    def testContainerIsPersistent(self):
        mgr = KeyManager()
        self.assertTrue(
            isinstance(mgr.__dict__["_SampleContainer__data"], PersistentMapping)
        )

    def testClear(self):
        self.mgr.clear()
        self.assertEqual(set(self.mgr["_system"]), {None})
        self.assertNotEqual(set(self.mgr["one"]), {None})
        self.assertNotEqual(set(self.mgr["two"]), {None})

    def testClearGivenRing(self):
        self.mgr.clear("one")
        self.assertNotEqual(set(self.mgr["_system"]), {None})
        self.assertEqual(set(self.mgr["one"]), {None})
        self.assertNotEqual(set(self.mgr["two"]), {None})

    def testClearAll(self):
        self.mgr.clear(None)
        self.assertEqual(set(self.mgr["_system"]), {None})
        self.assertEqual(set(self.mgr["one"]), {None})
        self.assertEqual(set(self.mgr["two"]), {None})

    def testClearUnknownRing(self):
        self.assertRaises(KeyError, self.mgr.clear, "missing")

    def testRotate(self):
        current_sys = self.mgr["_system"].current
        current_one = self.mgr["one"].current
        current_two = self.mgr["two"].current
        self.mgr.rotate()
        self.assertNotEqual(self.mgr["_system"].current, current_sys)
        self.assertEqual(self.mgr["_system"][1], current_sys)
        self.assertEqual(self.mgr["one"].current, current_one)
        self.assertEqual(self.mgr["one"][1], None)
        self.assertEqual(self.mgr["two"].current, current_two)
        self.assertEqual(self.mgr["two"][1], None)

    def testRotateGivenRing(self):
        current_sys = self.mgr["_system"].current
        current_one = self.mgr["one"].current
        current_two = self.mgr["two"].current
        self.mgr.rotate("one")
        self.assertEqual(self.mgr["_system"].current, current_sys)
        self.assertEqual(self.mgr["_system"][1], None)
        self.assertNotEqual(self.mgr["one"].current, current_one)
        self.assertEqual(self.mgr["one"][1], current_one)
        self.assertEqual(self.mgr["two"].current, current_two)
        self.assertEqual(self.mgr["two"][1], None)

    def testRotateAll(self):
        current_sys = self.mgr["_system"].current
        current_one = self.mgr["one"].current
        current_two = self.mgr["two"].current
        self.mgr.rotate(None)
        self.assertNotEqual(self.mgr["_system"].current, current_sys)
        self.assertEqual(self.mgr["_system"][1], current_sys)
        self.assertNotEqual(self.mgr["one"].current, current_one)
        self.assertEqual(self.mgr["one"][1], current_one)
        self.assertNotEqual(self.mgr["two"].current, current_two)
        self.assertEqual(self.mgr["two"][1], current_two)

    def testRotateUnknownRing(self):
        self.assertRaises(KeyError, self.mgr.clear, "missing")

    def testSecret(self):
        self.mgr["_system"][0] = marker
        self.assertTrue(self.mgr.secret() is marker)

    def testSecretGivenRing(self):
        self.mgr["one"][0] = marker
        self.assertTrue(self.mgr.secret("one") is marker)

    def testSecretUnknownRing(self):
        self.assertRaises(KeyError, self.mgr.secret, "missing")


def test_suite():
    suite = TestSuite()
    suite.addTest(defaultTestLoader.loadTestsFromTestCase(KeyManagerTests))
    return suite



from persistent.mapping import PersistentMapping
from plone.keyring.interfaces import IKeyManager
from plone.keyring.keyring import Keyring
from zope.container.sample import SampleContainer
from zope.interface import implementer


@implementer(IKeyManager)
class KeyManager(SampleContainer):
    def __init__(self, keyring_size=5):
        SampleContainer.__init__(self)

        if keyring_size < 1:
            keyring_size = 5  # prevent not having any keys

        self["_system"] = Keyring(keyring_size)
        self["_system"].fill()

        # to be used with anonymous users
        self["_anon"] = Keyring(keyring_size)
        self["_anon"].fill()

        # to be used with forms, plone.protect here..
        self["_forms"] = Keyring(keyring_size)
        self["_forms"].fill()

    def _newContainerData(self):
        return PersistentMapping()

    def clear(self, ring="_system"):
        if ring is None:
            for ring in self.values():
                ring.clear()
        else:
            self[ring].clear()

    def rotate(self, ring="_system"):
        if ring is None:
            for ring in self.values():
                ring.rotate()
        else:
            self[ring].rotate()

    def secret(self, ring="_system"):
        return self[ring].current



"""
Copyright (c) Django Software Foundation and individual contributors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

    3. Neither the name of Django nor the names of its contributors may be used
       to endorse or promote products derived from this software without
       specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from hashlib import sha256 as sha

import random
import time


try:
    random = random.SystemRandom()
    using_sysrandom = True
except NotImplementedError:
    using_sysrandom = False


# generated when process started, hard to guess
SECRET = random.randint(0, 1000000)


def get_random_string(
    length=12,
    allowed_chars="abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789",
):
    """
    Returns a securely generated random string.

    The default length of 12 with the a-z, A-Z, 0-9 character set returns
    a 71-bit value. log_2((26+26+10)^12) =~ 71 bits
    """
    if not using_sysrandom:
        # This is ugly, and a hack, but it makes things better than
        # the alternative of predictability. This re-seeds the PRNG
        # using a value that is hard for an attacker to predict, every
        # time a random string is required. This may change the
        # properties of the chosen random sequence slightly, but this
        # is better than absolute predictability.
        random.seed(sha(f"{random.getstate()}{time.time()}{SECRET}").digest())
    return "".join([random.choice(allowed_chars) for i in range(length)])
